#include <bsp.h>
/*
#include <bsp_level1.h>
*/
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <fcntl.h>
#include <math.h>
#include "mytime.h"

/* Constants and macros */
#define SAMPLES  1
#define MAXSIZES 50
#define INFINITY  999999999L
#define INITDIM 4
#define LASTDIM 8
#define DIMSTRIDE 4

#define ABS(x) (((x)>0)? (x) : -(x))
#define COORD(i,ix,iy) {ix = i/dimX; iy = i%dimX;}
#define distance(ix,iy,jx,jy) (ABS(jx-ix)+ABS(jy-iy))
#define graph(i,j) (ABS((j/dimX)-(i/dimX))+ABS((j%dimX)-(i%dimX)))
/* Types */
typedef int *pint;
typedef pint *grapht;
typedef struct {
  int source;
  int destine;
  int distance;
} TMin;

/* Prototypes */
grapht makeGraph( int nodes );
int *   prim_MST( int nodes, int ntasks, int mytid );
void   nearer( TMin *res, TMin *a, TMin *b, int *nbytes );
void   start();

/* Global variables */
int *d, *complement, *father;

struct timeval partialTime1[SAMPLES], partialTime2[SAMPLES];
   /* Computational Constants */
int Time_Size;
int Gi=0, step;
double SUM_INT_EF, SUM_INT_Fold, SUM_INT_HI;
struct timeval *Clock_C1, *Clock_C2;
struct timeval *Clock_D1, *Clock_D2;
struct timeval *Clock_Fold1, *Clock_Fold2;
struct timeval *Clock_E1, *Clock_E2;
							

double totalTime[MAXSIZES][SAMPLES];
double SUM;
/* Computational Constants */
double CD_Time[MAXSIZES][SAMPLES], EF_Time[MAXSIZES][SAMPLES], Fold_Time[MAXSIZES][SAMPLES], HI_Time[MAXSIZES][SAMPLES];
				

int    NNODES;
int    nprocs, mypid, dimX, dimY;
char  aux[255];
int m;

void writeTree(int *link) {
  int i;
	
	for(i=0;i<NNODES;i++)
		printf("%d: father of %d = (%d,%d) is %d\n",bsp_pid(),i,i/dimX,i%dimX,link[i]);
	printf("\n");
}

void init_all_mallocs() {
  if(!(Clock_C1= malloc((SAMPLES)*sizeof(struct timeval))))
    bsp_abort("not enough memory for array Clock_C1 \n");
  if(!(Clock_C2= malloc((SAMPLES)*sizeof(struct timeval))))
	  bsp_abort("not enough memory for array Clock_C2 \n");
  if(!(Clock_D1= malloc((Time_Size)*sizeof(struct timeval))))
	  bsp_abort("not enough memory for array Clock_D1 \n");
  if(!(Clock_D2= malloc((Time_Size)*sizeof(struct timeval))))
	  bsp_abort("not enough memory for array Clock_D2 \n");
  if(!(Clock_Fold1= malloc((Time_Size)*sizeof(struct timeval))))
	  bsp_abort("not enough memory for array Clock_Fold1 \n");
  if(!(Clock_Fold2= malloc((Time_Size)*sizeof(struct timeval))))
	  bsp_abort("not enough memory for array Clock_Fold2 \n");
  if(!(Clock_E1= malloc((Time_Size)*sizeof(struct timeval))))
	  bsp_abort("not enough memory for array Clock_E1 \n");
  if(!(Clock_E2= malloc((Time_Size)*sizeof(struct timeval))))
	  bsp_abort("not enough memory for array Clock_E2 \n");

  if(!(d = malloc((NNODES)*sizeof(int))))
		bsp_abort("not enough memory for array d in prim_MST\n");
  if(!(complement = malloc((NNODES)*sizeof(int))))
		bsp_abort("not enough memory for array complement in prim_MST\n");
  if(!(father = malloc((NNODES)*sizeof(int))))
		bsp_abort("not enough memory for array father in prim_MST\n");
}

void free_all_mallocs() {
  free(Clock_C1);
		free(Clock_C2);
		free(Clock_D1);
		free(Clock_D2);
		free(Clock_Fold1);
		free(Clock_Fold2);
		free(Clock_E1);
		free(Clock_E2);
}

void start(void) {
  int   *tree, *link, i, j, k;

  bsp_begin(bsp_nprocs());

  nprocs = bsp_nprocs();
  mypid  = bsp_pid();

  for (k=0, dimY = INITDIM, dimX = INITDIM; dimX <= LASTDIM; k++, dimX += DIMSTRIDE, dimY += DIMSTRIDE) {
		NNODES = dimX*dimY;
		Time_Size = NNODES-1;

    init_all_mallocs();

		for (m = 0; m < SAMPLES; m++) {
				bsp_sync();
				/* ------------ PRIM --------------- */
				GETTIME(partialTime1[m]);
				
				link = prim_MST(NNODES, nprocs, mypid);

				GETTIME(partialTime2[m]);

				DIFTIME(partialTime2[m], partialTime1[m], totalTime[k][m]);
				
				/* COMPUTE INTERNAL LOOP COMPUTATIONAL CONSTANTS  */
				DIFTIME(Clock_C2[m], Clock_C1[m], CD_Time[k][m]);

				if (bsp_pid() == 0) {
					SUM_INT_EF   = 0.0;
					SUM_INT_Fold = 0.0;
					SUM_INT_HI   = 0.0;
				  for (j=0; j<Time_Size; j++) {
					  DIFTIME(Clock_D2[j], Clock_D1[j], SUM);
						SUM_INT_EF   += SUM;
            DIFTIME(Clock_Fold2[j], Clock_Fold1[j], SUM);
						SUM_INT_Fold += SUM;
						DIFTIME(Clock_E2[j], Clock_E1[j], SUM);
						SUM_INT_HI   += SUM;
					}
				}

				EF_Time[k][m]   = SUM_INT_EF/Time_Size;
				Fold_Time[k][m] = SUM_INT_Fold/Time_Size;
				HI_Time[k][m]   = SUM_INT_HI/Time_Size;

				writeTree(link);

				if (!checkMeshSolution(link,NNODES)) 
					printf("\n[%d]: ERROR!!!!!!!!!!!!!!!!!!!\n", bsp_pid());
			}
			free_all_mallocs();
  }


		if (bsp_pid() == 0) {
				for (k=0, dimY = INITDIM, dimX = INITDIM; dimX <= LASTDIM; k++, dimX += DIMSTRIDE, dimY += DIMSTRIDE) {
				  printf("\n        NAME       Size       Rep      NUMPROCS      Par_Time[rep]    CD       EF         Fold_Time        HI ");

				  for(m=0;m<SAMPLES;m++) {
            printf("\n %10d %10d  %10d %10d  %12.8f  %12.8f %12.8f %12.8f %12.8f ",
						bsp_pid(), dimX*dimY, m, nprocs, totalTime[k][m], CD_Time[k][m],
						EF_Time[k][m], Fold_Time[k][m], HI_Time[k][m]);
			  	}
		  	}
		}

  bsp_end();
}

int checkMeshSolution(int *father, int nodes) {
	int i, fx,fy,ix,iy, succes;

  succes = 1;
  for (i=1;i<nodes;i++) {
		COORD(father[i],fx,fy);
	  COORD(i,ix,iy);
		if (distance(ix,iy,fx,fy) != 1) { 
			printf("[%d] Error. Father of %d (%d,%d) is %d (%d,%d)",
			bsp_pid(),i,ix,iy,father[i],fx,fy);
			succes = 0;
			return succes;
		}
	}
	return succes;
}

int main( int argc, char **argv ) {

  start();
  return 0;
}

int * prim_MST( int nodes, int ntasks, int mytid ) {
  int  step, i, new, temp;
  TMin min_local, min_global;
	int min_distance, nearest_point, contact;
	
	GETTIME(Clock_C1[m]);   /* ------------TIME --------- */

	for (i = mytid; i < nodes; i+= ntasks) {
		complement[i] = 1; /* any others are in the tree complement */
		d[i] = graph(0,i); 
    father[i] = 0;
		/* at any time holds: d[j] = dist(j,tree) for j in complement */
	}
	father[0] = 0;
  complement[0] = 0; /* 0 is the root of the minimum spanning tree */

	GETTIME(Clock_C2[m]);   /* ------------TIME --------- */

  for (step = 0; step < (nodes-1); step++) { /*this loop can be improved */

	  GETTIME(Clock_D1[step]);   /* ------------TIME --------- */

    min_distance = INFINITY;
    for (i = mytid; i < nodes; i+= ntasks) {
			if (complement[i]) {
				if (d[i] < min_distance) {
					min_distance = d[i];
					contact = father[i];
					nearest_point = i;
				}
			}
		}

		min_local.source = contact;
		min_local.destine = nearest_point;
		min_local.distance = min_distance;

		GETTIME(Clock_D2[step]);   /* ------------TIME --------- */

		/* global reduction to obtain the nearest point */

    GETTIME(Clock_Fold1[step]);   /* ------------TIME --------- */

		bsp_fold((void (*)(void*,void*,void*,int*))nearer, 
									 &min_local, &min_global, sizeof(TMin));

    GETTIME(Clock_Fold2[step]);   /* ------------TIME --------- */



		/* improvement: it would be better to keep all the elements */
		/* reaching minimum distance and insert them in only one step */

    GETTIME(Clock_E1[step]);   /* ------------TIME --------- */

		new = min_global.destine;
    complement[new] = 0; /* this element belongs now to the tree */
		father[new] = min_global.source;

    for (i = mytid; i < nodes; i+= ntasks) {
			if (complement[i]) {
				temp = graph(i,new); 
				if (d[i] > temp) {
					d[i] = temp;
					father[i] = new;
				}
			}
		} /* for (i = 0 ... update d[i] */
		GETTIME(Clock_E2[step]);   /* ------------TIME --------- */
	} /* step loop */
	return father;
}

void nearer( TMin *res, TMin *a, TMin *b, int *nbytes ) {

  if (a->distance < b->distance) {
    res->distance = a->distance;
    res->source   = a->source;
    res->destine  = a->destine;
    return;
  }  
  if (a->distance == b->distance) {
    if (a->source < b->source) {
      res->distance = a->distance;
      res->source   = a->source;
      res->destine  = a->destine;
      return;
    }
    if ((a->source == b->source) && (a->destine < b->destine)) {
      res->distance = a->distance;
      res->source   = a->source;
      res->destine  = a->destine;
      return;
    }
  }
  res->distance = b->distance;
  res->source   = b->source;
  res->destine  = b->destine;
 }
